@@ -3,14 +3,16 @@ import sys
 import json
 import math
 import torch
+import hashlib
+import inspect
 import argparse
-import torchvision
 import urllib.request
 
 import numpy as np
 import torch.nn as nn
 import torch.nn.functional as F
 
+from abc import ABCMeta
 from PIL import Image
 from PIL import ImageDraw
 from torch import Tensor
@@ -29,22 +31,23 @@ from collections import defaultdict
 from collections import OrderedDict
 from torch.optim import Optimizer
 from cftool.misc import prod
+from cftool.misc import print_info
+from cftool.misc import print_warning
 from cftool.misc import check_requires
 from cftool.misc import shallow_copy_dict
-from cftool.misc import context_error_handler
+from cftool.misc import truncate_string_to_length
 from cftool.misc import DownloadProgressBar
-from cftool.array import arr_type
+from cftool.array import to_torch
 from cftool.array import to_standard
+from cftool.types import np_dict_type
+from cftool.types import tensor_dict_type
 
 from ..types import data_type
 from ..types import param_type
-from ..types import np_dict_type
-from ..types import tensor_dict_type
 from ..types import sample_weights_type
 from ..constants import CACHE_DIR
 from ..constants import INPUT_KEY
-from ..constants import INFO_PREFIX
-from ..constants import WARNING_PREFIX
+from ..constants import WORKPLACE_ENVIRON_KEY
 
 try:
     import matplotlib.pyplot as plt
@@ -63,16 +66,12 @@ except:
 # general
 
 
-def filter_kw(
-    fn: Callable,
-    kwargs: Dict[str, Any],
-    strict: bool = True,
-) -> Dict[str, Any]:
-    kw = {}
-    for k, v in kwargs.items():
-        if check_requires(fn, k, strict):
-            kw[k] = v
-    return kw
+def _get_environ_workplace() -> Optional[str]:
+    return os.environ.get(WORKPLACE_ENVIRON_KEY)
+
+
+def _set_environ_workplace(workplace: str) -> None:
+    os.environ[WORKPLACE_ENVIRON_KEY] = workplace
 
 
 def check_is_ci() -> bool:
@@ -82,12 +81,30 @@ def check_is_ci() -> bool:
     return bool(args.ci)
 
 
-def check_available(tag: str, repo: str, name: str) -> bool:
+class FileInfo(NamedTuple):
+    sha: str
+    st_size: int
+    download_url: Optional[str] = None
+
+
+def check_available(tag: str, repo: str, name: str) -> Optional[FileInfo]:
     with open(os.path.join(os.path.dirname(__file__), "available.json"), "r") as f:
         available = json.load(f)
     if repo != "pretrained-models":
-        return True
-    return name in available[tag]
+        return FileInfo("", 0)
+    info = available[tag].get(name)
+    return None if info is None else FileInfo(**info)
+
+
+def _get_file_info(path: str) -> FileInfo:
+    with open(path, "rb") as f:
+        sha = hashlib.sha256(f.read()).hexdigest()
+    return FileInfo(sha, os.stat(path).st_size)
+
+
+def _check_sha(path: str, tgt_sha: str) -> bool:
+    with open(path, "rb") as f:
+        return hashlib.sha256(f.read()).hexdigest() == tgt_sha
 
 
 def download(
@@ -96,135 +113,79 @@ def download(
     name: str,
     root: str,
     extension: str,
-    force_download: bool,
-    remove_zip: Optional[bool],
-    extract_zip: bool,
+    *,
+    check_sha: bool = False,
+    remove_zip: bool = True,
 ) -> str:
-    if not check_available(tag, repo, name):
+    info = check_available(tag, repo, name)
+    if info is None:
         raise ValueError(f"'{name}' is currently not available at '{tag}'")
-    prefix = f"https://github.com/carefree0910/{repo}/releases/download/{tag}/"
     os.makedirs(root, exist_ok=True)
     file = f"{name}.{extension}"
     path = os.path.join(root, file)
     is_zip = extension == "zip"
     zip_folder_path = os.path.join(root, name)
-    exists = os.path.isdir(zip_folder_path) if is_zip else os.path.isfile(path)
-    if exists and not force_download:
-        return zip_folder_path if is_zip else path
+    if is_zip and os.path.isdir(zip_folder_path):
+        return zip_folder_path
+    fmt = "cache file is detected but {}, it will be re-downloaded"
+    if not is_zip and os.path.isfile(path):
+        if os.stat(path).st_size != info.st_size:
+            print_warning(fmt.format("st_size is not correct"))
+        else:
+            if not check_sha or _check_sha(path, info.sha):
+                return path
+            print_warning(fmt.format("sha is not correct"))
     with DownloadProgressBar(unit="B", unit_scale=True, miniters=1, desc=name) as t:
+        if info.download_url is not None:
+            url = info.download_url
+        else:
+            prefix = f"https://github.com/carefree0910/{repo}/releases/download/{tag}/"
+            url = f"{prefix}{file}"
         urllib.request.urlretrieve(
-            f"{prefix}{file}",
+            url,
             filename=path,
             reporthook=t.update_to,
         )
     if not is_zip:
         return path
-    if extract_zip:
-        with ZipFile(path, "r") as zip_ref:
-            zip_ref.extractall(zip_folder_path)
-    if remove_zip is None:
-        remove_zip = extract_zip
+    with ZipFile(path, "r") as zip_ref:
+        zip_ref.extractall(zip_folder_path)
     if remove_zip:
-        if extract_zip:
-            os.remove(path)
-        else:
-            print(
-                f"{WARNING_PREFIX}zip file is not extracted, "
-                f"so we will not remove it!"
-            )
+        os.remove(path)
     return zip_folder_path
 
 
-def download_data_info(
-    name: str,
-    *,
-    root: str = os.path.join(CACHE_DIR, "data_info"),
-    force_download: bool = False,
-) -> str:
-    return download(
-        "data_info",
-        "pretrained-models",
-        name,
-        root,
-        "json",
-        force_download,
-        None,
-        False,
-    )
-
-
 def download_tokenizer(
     name: str,
     *,
     root: str = os.path.join(CACHE_DIR, "tokenizers"),
-    force_download: bool = False,
 ) -> str:
-    return download(
-        "tokenizers",
-        "pretrained-models",
-        name,
-        root,
-        "pkl",
-        force_download,
-        None,
-        False,
-    )
+    return download("tokenizers", "pretrained-models", name, root, "pkl")
 
 
 def download_model(
     name: str,
     *,
     root: str = os.path.join(CACHE_DIR, "models"),
-    force_download: bool = False,
 ) -> str:
-    return download(
-        "checkpoints",
-        "pretrained-models",
-        name,
-        root,
-        "pt",
-        force_download,
-        None,
-        False,
-    )
+    return download("checkpoints", "pretrained-models", name, root, "pt")
 
 
 def download_reference(
     name: str,
     *,
     root: str = os.path.join(CACHE_DIR, "reference"),
-    force_download: bool = False,
 ) -> str:
-    return download(
-        "reference",
-        "pretrained-models",
-        name,
-        root,
-        "pt",
-        force_download,
-        None,
-        False,
-    )
+    return download("reference", "pretrained-models", name, root, "pt")
 
 
 def download_dataset(
     name: str,
     *,
     root: str = os.getcwd(),
-    force_download: bool = False,
-    remove_zip: Optional[bool] = None,
-    extract_zip: bool = True,
+    remove_zip: bool = True,
 ) -> str:
-    return download(
-        "latest",
-        "datasets",
-        name,
-        root,
-        "zip",
-        force_download,
-        remove_zip,
-        extract_zip,
-    )
+    return download("latest", "datasets", name, root, "zip", remove_zip=remove_zip)
 
 
 def get_compatible_name(
@@ -260,13 +221,40 @@ def get_compatible_name(
         if check_available(tag, repo, compatible_name):
             name = compatible_name
         else:
-            print(
-                f"{WARNING_PREFIX}compatible name '{compatible_name}' is not available "
+            print_warning(
+                f"compatible name '{compatible_name}' is not available "
                 f"on the server, will use the original name ({name}) instead"
             )
     return name
 
 
+class ConfigMeta(ABCMeta):
+    def __new__(mcs, *args: Any, **kwargs: Any) -> type:
+        name, bases, attr = args[:3]
+        original_init = attr["__init__"]
+
+        def __init__(self: Any, *args_: Any, **kwargs_: Any) -> None:
+            if getattr(self, "_in_control_", False):
+                original_init(self, *args_, **kwargs_)
+            else:
+                self._is_root_ = True
+                self._in_control_ = True
+                self.config = shallow_copy_dict(kwargs_)
+                signatures = list(inspect.signature(original_init).parameters.items())
+                for arg, (k, _) in zip(args_, signatures[1:]):
+                    if not isinstance(arg, dict):
+                        self.config[k] = arg
+                    else:
+                        self.config[k] = shallow_copy_dict(arg)
+                original_init(self, *args_, **kwargs_)
+            if getattr(self, "_is_root_", False):
+                del self._is_root_
+                del self._in_control_
+
+        attr["__init__"] = __init__
+        return type(name, bases, attr)
+
+
 class WeightsStrategy:
     def __init__(self, strategy: Optional[str]):
         self.strategy = strategy
@@ -291,9 +279,9 @@ class WeightsStrategy:
 
     def visualize(self, export_path: str = "weights_strategy.png") -> None:
         if plt is None:
-            raise ValueError(f"{WARNING_PREFIX}`matplotlib` is needed for `visualize`")
+            raise ValueError("`matplotlib` is needed for `visualize`")
         if show_or_save is None:
-            raise ValueError(f"{WARNING_PREFIX}`carefree-ml` is needed for `visualize`")
+            raise ValueError("`carefree-ml` is needed for `visualize`")
         n = 1000
         x = np.linspace(0, 1, n)
         y = self(n, 0)
@@ -307,21 +295,6 @@ class WeightsStrategy:
 # dl
 
 
-def get_label_predictions(logits: np.ndarray, threshold: float) -> np.ndarray:
-    # binary classification
-    if logits.shape[-1] == 1:
-        logit_threshold = math.log(threshold / (1.0 - threshold))
-        return (logits > logit_threshold).astype(int)
-    return logits.argmax(1)[..., None]
-
-
-def get_full_logits(logits: np.ndarray) -> np.ndarray:
-    # binary classification
-    if logits.shape[-1] == 1:
-        logits = np.concatenate([-logits, logits], axis=-1)
-    return logits
-
-
 def inject_debug(config: Dict[str, Any]) -> None:
     config["fixed_steps"] = 1
     config["valid_portion"] = 1.0e-4
@@ -345,41 +318,10 @@ def fix_denormal_states(
         if num_denormal > 0:
             new_states[k][denormal] = v.new_zeros(num_denormal)
     if verbose:
-        print(f"{INFO_PREFIX}denormal ratio : {num_denormal_total / num_total:8.6f}")
+        print_info(f"denormal ratio : {num_denormal_total / num_total:8.6f}")
     return new_states
 
 
-def toggle_optimizer(m: nn.Module, optimizer: Optimizer) -> None:
-    for param in m.parameters():
-        param.requires_grad = False
-    for group in optimizer.param_groups:
-        for param in group["params"]:
-            param.requires_grad = True
-
-
-def to_torch(arr: np.ndarray) -> Tensor:
-    return torch.from_numpy(to_standard(arr))
-
-
-def to_numpy(tensor: Tensor) -> np.ndarray:
-    return tensor.detach().cpu().numpy()
-
-
-def to_device(
-    batch: tensor_dict_type,
-    device: torch.device,
-    **kwargs: Any,
-) -> tensor_dict_type:
-    return {
-        k: v.to(device, **kwargs)
-        if isinstance(v, Tensor)
-        else [vv.to(device, **kwargs) if isinstance(vv, Tensor) else vv for vv in v]
-        if isinstance(v, list)
-        else v
-        for k, v in batch.items()
-    }
-
-
 def has_batch_norms(m: nn.Module) -> bool:
     bn_types = (nn.BatchNorm1d, nn.BatchNorm2d, nn.BatchNorm3d, nn.SyncBatchNorm)
     for name, module in m.named_modules():
@@ -464,6 +406,14 @@ def set_requires_grad(module: nn.Module, requires_grad: bool = False) -> None:
         param.requires_grad = requires_grad
 
 
+def freeze(module: nn.Module) -> nn.Module:
+    module.eval()
+    # make sure that the module will never go back to `train` mode
+    module.train = lambda m, mode=True: m  # type: ignore
+    set_requires_grad(module, False)
+    return module
+
+
 def scheduler_requires_metric(scheduler: Any) -> bool:
     return check_requires(scheduler.step, "metrics")
 
@@ -479,41 +429,44 @@ def summary(
     def _get_param_counts(module_: nn.Module) -> Tuple[int, int]:
         num_params = 0
         num_trainable_params = 0
-        for param in module_.parameters():
-            local_num_params = int(round(prod(param.data.shape)))
+        for p in module_.parameters():
+            local_num_params = int(round(prod(p.data.shape)))
             num_params += local_num_params
-            if param.requires_grad:
+            if p.requires_grad:
                 num_trainable_params += local_num_params
         return num_params, num_trainable_params
 
     def register_hook(module: nn.Module) -> None:
+        def inject_output_shape(output: Any, res: Dict[int, Any]) -> None:
+            idx = 0 if not res else max(res)
+            if isinstance(output, Tensor):
+                o_shape = list(output.shape)
+                if o_shape:
+                    o_shape[0] = -1
+                res[idx + 1] = o_shape
+                return
+            if isinstance(output, (list, tuple)):
+                o_res = res[idx + 1] = {}
+                for o in output:
+                    inject_output_shape(o, o_res)
+
         def hook(module_: nn.Module, inp: Any, output: Any) -> None:
             m_name = module_names.get(module_)
             if m_name is None:
                 return
 
+            if not inp:
+                return
             inp = inp[0]
             if not isinstance(inp, Tensor):
                 return
-            if isinstance(output, (list, tuple)):
-                for element in output:
-                    if not isinstance(element, Tensor):
-                        return
-            elif not isinstance(output, Tensor):
-                return
 
             m_dict: OrderedDict[str, Any] = OrderedDict()
-            m_dict["input_shape"] = list(inp.size())
+            m_dict["input_shape"] = list(inp.shape)
             if len(m_dict["input_shape"]) > 0:
                 m_dict["input_shape"][0] = -1
-            if isinstance(output, (list, tuple)):
-                m_dict["output_shape"] = [[-1] + list(o.size())[1:] for o in output]
-                m_dict["is_multiple_output"] = True
-            else:
-                m_dict["output_shape"] = list(output.size())
-                if len(m_dict["output_shape"]) > 0:
-                    m_dict["output_shape"][0] = -1
-                m_dict["is_multiple_output"] = False
+            output_shape_res = m_dict["output_shape"] = {}
+            inject_output_shape(output, output_shape_res)
 
             num_params_, num_trainable_params_ = _get_param_counts(module_)
             m_dict["num_params"] = num_params_
@@ -614,28 +567,49 @@ def summary(
     messages = ["=" * line_length]
     line_format = "{:30}  {:>20} {:>40} {:>20}"
     headers = "Layer (type)", "Input Shape", "Output Shape", "Trainable Param #"
-    line_new = line_format.format(*headers)
-    messages.append(line_new)
+    messages.append(line_format.format(*headers))
     messages.append("-" * line_length)
     total_output = 0
     for layer, layer_summary in summary_dict.items():
         layer_name = "-".join(layer.split("-")[:-1])
+        valid_layer_name = layer_name.strip()
+        num_spaces = len(layer_name) - len(valid_layer_name)
+        valid_layer_name = truncate_string_to_length(valid_layer_name, 30 - num_spaces)
+        layer_name = " " * num_spaces + valid_layer_name
         if layer_summary is None:
-            line_new = line_format.format(layer_name, "", "", "")
+            messages.append(line_format.format(layer_name, "", "", ""))
         else:
-            line_new = line_format.format(
-                layer_name,
-                str(layer_summary["input_shape"]),
-                str(layer_summary["output_shape"]),
-                "{0:,}".format(layer_summary["num_trainable_params"]),
-            )
-            output_shape = layer_summary["output_shape"]
-            is_multiple_output = layer_summary["is_multiple_output"]
-            if not is_multiple_output:
-                output_shape = [output_shape]
-            for shape in output_shape:
+            is_title = True
+            all_output_shapes: List[List[int]] = []
+
+            def _inject(output_shape_item: Dict[int, Any], prefix: str) -> None:
+                only_one = len(output_shape_item) == 1
+                for i, idx in enumerate(sorted(output_shape_item)):
+                    if not prefix and only_one:
+                        idx_prefix = ""
+                    else:
+                        idx_prefix = f"{prefix}{idx}."
+                    value = output_shape_item[idx]
+                    if isinstance(value, dict):
+                        _inject(value, idx_prefix)
+                        continue
+                    output_shape_str = f"{idx_prefix} {str(value):>16s}"
+                    ntp_str = "{0:,}".format(layer_summary["num_trainable_params"])
+                    nonlocal is_title
+                    messages.append(
+                        line_format.format(
+                            layer_name if is_title else "",
+                            str(layer_summary["input_shape"]) if is_title else "",
+                            output_shape_str,
+                            ntp_str if is_title else "",
+                        )
+                    )
+                    is_title = False
+                    all_output_shapes.append(value)
+
+            _inject(layer_summary["output_shape"], "")
+            for shape in all_output_shapes:
                 total_output += prod(shape)
-        messages.append(line_new)
 
     total_params, trainable_params = _get_param_counts(model)
     # assume 4 bytes/number (float on cuda).
@@ -686,7 +660,49 @@ def get_world_size() -> int:
     return 1 if ddp_info is None else ddp_info.world_size
 
 
-class mode_context(context_error_handler):
+class toggle_optimizer:
+    """
+    Help focusing gradients on specific optimizer and recovering previous states
+
+    This is a context controller for requiring and only requiring grads for parameters
+    of the given optimizer at the beginning, and back to previous grads requiring states
+    at the end.
+
+    Examples
+    --------
+    >>> module = nn.Module()
+    >>> optimizer = torch.optim.Adam()
+    >>> with toggle_optimizer(module, optimizer):
+    >>>     pass  # do something
+
+    """
+
+    def __init__(self, m: nn.Module, optimizer: Optimizer, *, enabled: bool = True):
+        self.m = m
+        self.optimizer = optimizer
+        self.enabled = enabled
+        self.requires_grad: Dict[str, bool] = {}
+
+    def __enter__(self) -> None:
+        if not self.enabled:
+            return
+        self.requires_grad = {k: p.requires_grad for k, p in self.m.named_parameters()}
+        for p in self.m.parameters():
+            p.requires_grad = False
+        for group in self.optimizer.param_groups:
+            for p in group["params"]:
+                p.requires_grad = True
+
+    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
+        if not self.enabled:
+            return
+        for k, p in self.m.named_parameters():
+            requires_grad = self.requires_grad.get(k)
+            if requires_grad is not None:
+                p.requires_grad = requires_grad
+
+
+class mode_context:
     """
     Help entering specific mode and recovering previous mode
 
@@ -778,6 +794,22 @@ class eval_context(mode_context):
         )
 
 
+class no_grad_context(torch.no_grad):
+    def __init__(self, *, enabled: bool):
+        super().__init__()
+        self.enabled = enabled
+
+    def __enter__(self) -> None:
+        if not self.enabled:
+            return
+        super().__enter__()
+
+    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
+        if not self.enabled:
+            return
+        super().__exit__(exc_type, exc_val, exc_tb)
+
+
 class Initializer:
     """
     Initializer for neural network weights
@@ -811,7 +843,7 @@ class Initializer:
     @classmethod
     def add_initializer(cls, f: Callable, name: str) -> None:
         if name in cls.defined_initialization:
-            print(f"{WARNING_PREFIX}'{name}' initializer is already defined")
+            print_warning(f"'{name}' initializer is already defined")
             return
         cls.defined_initialization.add(name)
         cls.custom_initializer[name] = f
@@ -850,8 +882,8 @@ class Initializer:
                 weight_base[invalid] = param.new_empty(num_invalid).normal_()
                 invalid = get_invalid(weight_base)
         if not success:
-            print(
-                f"{WARNING_PREFIX}invalid ratio for truncated normal : "
+            print_warning(
+                "invalid ratio for truncated normal : "
                 f"{invalid.to(torch.float32).mean():8.6f}, it might cause by "
                 f"too little epoch ({epoch}) or too small tolerance ({tol})",
             )
@@ -874,10 +906,12 @@ class DropNoGradStatesMixin:
     ) -> tensor_dict_type:
         states = super().state_dict(destination=destination, prefix=prefix, keep_vars=keep_vars)  # type: ignore
         for key, _ in self.named_buffers():  # type: ignore
-            states.pop(key)
+            if states.pop(key, None) is None:
+                states.pop(f"core.{key}")
         for key, value in self.named_parameters():  # type: ignore
             if not value.requires_grad:
-                states.pop(key)
+                if states.pop(key, None) is None:
+                    states.pop(f"core.{key}")
         return states
 
     def load_state_dict(
@@ -912,6 +946,43 @@ class ONNX:
         return dict(zip(self.output_names, self.ort_session.run(None, ort_inputs)))
 
 
+def gradient_checkpoint(func: Callable, inputs: Any, params: Any, enabled: bool) -> Any:
+    if not enabled:
+        return func(*inputs)
+    args = tuple(inputs) + tuple(params)
+    return GradientCheckpointFunction.apply(func, len(inputs), *args)
+
+
+class GradientCheckpointFunction(torch.autograd.Function):
+    @staticmethod
+    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Any:
+        run_function, length, *args = args  # type: ignore
+        ctx.run_function = run_function
+        ctx.input_tensors = list(args[:length])
+        ctx.input_params = list(args[length:])
+
+        with torch.no_grad():
+            output_tensors = ctx.run_function(*ctx.input_tensors)
+        return output_tensors
+
+    @staticmethod
+    def backward(ctx: Any, *grad_outputs: Any) -> Any:
+        ctx.input_tensors = [x.detach().requires_grad_(True) for x in ctx.input_tensors]
+        with torch.enable_grad():
+            shallow_copies = [x.view_as(x) for x in ctx.input_tensors]
+            output_tensors = ctx.run_function(*shallow_copies)
+        input_grads = torch.autograd.grad(
+            output_tensors,
+            ctx.input_tensors + ctx.input_params,
+            grad_outputs,
+            allow_unused=True,
+        )
+        del ctx.input_tensors
+        del ctx.input_params
+        del output_tensors
+        return (None, None) + input_grads
+
+
 # ml
 
 
@@ -977,9 +1048,9 @@ def interpolate(
     if factor is not None:
         template = "`{}` will take no affect because `factor` is provided"
         if size is not None:
-            print(f"{WARNING_PREFIX}{template.format('size')}")
+            print_warning(template.format("size"))
         if anchor is not None:
-            print(f"{WARNING_PREFIX}{template.format('anchor')}")
+            print_warning(template.format("anchor"))
         if factor == 1.0 or factor == (1.0, 1.0):
             return src
         if not determinate:
@@ -1042,22 +1113,6 @@ def adain_with_tensor(src: Tensor, tgt: Tensor, *, determinate: bool = False) ->
     return adain_with_params(src, tgt_mean, tgt_std, determinate=determinate)
 
 
-def make_grid(arr: arr_type, n_row: Optional[int] = None) -> Tensor:
-    if isinstance(arr, np.ndarray):
-        arr = to_torch(arr)
-    if n_row is None:
-        n_row = math.ceil(math.sqrt(len(arr)))
-    return torchvision.utils.make_grid(arr, n_row)
-
-
-def save_images(arr: arr_type, path: str, n_row: Optional[int] = None) -> None:
-    if isinstance(arr, np.ndarray):
-        arr = to_torch(arr)
-    if n_row is None:
-        n_row = math.ceil(math.sqrt(len(arr)))
-    torchvision.utils.save_image(arr, path, normalize=True, nrow=n_row)
-
-
 def make_indices_visualization_map(indices: Tensor) -> Tensor:
     images = []
     for idx in indices.view(-1).tolist():
