@@ -1,9 +1,11 @@
+from .ae import *
 from .gan import *
 from .ssl import *
 from .vae import *
 from .encoder import *
 from .decoder import *
 from .stylizer import *
+from .diffusion import *
 from .generator import *
 from .segmentor import *
 from .classifier import *
