@@ -10,24 +10,29 @@ from typing import Any
 from typing import Dict
 from typing import Tuple
 from typing import Optional
+from torch.optim import Optimizer
 from cftool.misc import update_dict
 from cftool.misc import shallow_copy_dict
+from cftool.array import to_device
 from cftool.array import l2_normalize
+from cftool.types import tensor_dict_type
 from torch.nn.parallel import DistributedDataParallel as DDP
+from torch.cuda.amp import GradScaler
 
-from ..encoder import Encoder1DBase
+from ..encoder import run_encoder
+from ..encoder import Encoder1DMixin
 from ....data import CVLoader
-from ....types import tensor_dict_type
+from ....protocol import ITrainer
 from ....protocol import StepOutputs
 from ....protocol import TrainerState
 from ....protocol import MetricsOutputs
-from ....protocol import ModelWithCustomSteps
 from ....constants import LOSS_KEY
 from ....constants import INPUT_KEY
 from ....constants import LATENT_KEY
-from ....misc.toolkit import to_device
 from ....misc.toolkit import get_world_size
 from ....misc.toolkit import has_batch_norms
+from ....misc.internal_.register import register_custom_module
+from ....misc.internal_.register import CustomModule
 
 
 def _get_dino_defaults(name: str) -> Dict[str, Any]:
@@ -66,7 +71,7 @@ def cosine_scheduler(
 
 
 class MultiCropWrapper(nn.Module):
-    def __init__(self, backbone: nn.Module, head: nn.Module):
+    def __init__(self, backbone: Encoder1DMixin, head: nn.Module):
         super().__init__()
         backbone.fc, backbone.head = nn.Identity(), nn.Identity()
         self.backbone = backbone
@@ -98,7 +103,7 @@ class MultiCropWrapper(nn.Module):
         for end_idx in idx_crops:
             local_batch = shallow_copy_dict(batch)
             local_batch[INPUT_KEY] = torch.cat(img_crops[start_idx:end_idx])
-            idx_rs = self.backbone(batch_idx, local_batch, state, **kwargs)
+            idx_rs = run_encoder(self.backbone, batch_idx, local_batch, state, **kwargs)
             idx_out = idx_rs[LATENT_KEY]
             if isinstance(idx_out, tuple):
                 idx_out = idx_out[0]
@@ -243,11 +248,12 @@ class DINOEvaluateLoss:
         return loss.item()
 
 
-@ModelWithCustomSteps.register("dino")
-class DINO(ModelWithCustomSteps):
-    custom_params_groups = True
-    custom_ddp_initialization = True
-
+@register_custom_module(
+    "dino",
+    custom_params_groups=True,
+    custom_ddp_initialization=True,
+)
+class DINO(CustomModule):
     lr_schedule: Optional[Scheduler]
     wd_schedule: Optional[Scheduler]
     momentum_schedule: Optional[Scheduler]
@@ -276,8 +282,8 @@ class DINO(ModelWithCustomSteps):
         base = update_dict(encoder1d_config or {}, _get_dino_defaults(encoder1d))
         student_cfg = update_dict(student_specific or {}, shallow_copy_dict(base))
         teacher_cfg = update_dict(teacher_specific or {}, shallow_copy_dict(base))
-        student = Encoder1DBase.make(encoder1d, student_cfg)
-        teacher = Encoder1DBase.make(encoder1d, teacher_cfg)
+        student = Encoder1DMixin.make(encoder1d, student_cfg)
+        teacher = Encoder1DMixin.make(encoder1d, teacher_cfg)
         self.ddp_student = self.ddp_teacher = None
         self.student = MultiCropWrapper(
             student,
@@ -326,7 +332,8 @@ class DINO(ModelWithCustomSteps):
         state: Optional[TrainerState] = None,
         **kwargs: Any,
     ) -> tensor_dict_type:
-        net = self.student.backbone(batch_idx, batch, state, **kwargs)[LATENT_KEY]
+        res = run_encoder(self.student.backbone, batch_idx, batch, state, **kwargs)
+        net = res[LATENT_KEY]
         net = l2_normalize(net)
         return {LATENT_KEY: net}
 
@@ -335,6 +342,9 @@ class DINO(ModelWithCustomSteps):
         net = self.get_latent(inp, determinate=True)
         return net.view(inp.shape[0], self.student.backbone.latent_dim)
 
+    def summary_forward(self, batch_idx: int, batch: tensor_dict_type) -> None:
+        self.student(batch_idx, to_device(batch, self.device))
+
     def get_latent(self, net: Tensor, **kwargs: Any) -> Tensor:
         return self.forward(0, {INPUT_KEY: net}, **kwargs)[LATENT_KEY]
 
@@ -353,27 +363,24 @@ class DINO(ModelWithCustomSteps):
                 states.pop(k)
         return states
 
-    def summary_forward(self, batch_idx: int, batch: tensor_dict_type) -> None:
-        self.student(batch_idx, to_device(batch, self.device))
-
     def _get_outputs(
         self,
         batch_idx: int,
         batch: tensor_dict_type,
-        trainer: Any,
+        state: TrainerState,
         forward_kwargs: Dict[str, Any],
     ) -> tensor_dict_type:
         teacher_output = self.teacher_for_training(
             batch_idx,
             batch,
-            trainer.state,
+            state,
             img_end_idx=2,
             **forward_kwargs,
         )
         student_output = self.student_for_training(
             batch_idx,
             batch,
-            trainer.state,
+            state,
             **forward_kwargs,
         )
         return {"student": student_output, "teacher": teacher_output}
@@ -382,27 +389,30 @@ class DINO(ModelWithCustomSteps):
         self,
         batch_idx: int,
         batch: tensor_dict_type,
-        trainer: Any,
+        state: TrainerState,
+        use_amp: bool,
         forward_kwargs: Dict[str, Any],
     ) -> Tuple[tensor_dict_type, Tensor]:
-        with torch.cuda.amp.autocast(enabled=trainer.use_amp):
-            outputs = self._get_outputs(batch_idx, batch, trainer, forward_kwargs)
-            epoch = trainer.state.epoch
+        with torch.cuda.amp.autocast(enabled=use_amp):
+            outputs = self._get_outputs(batch_idx, batch, state, forward_kwargs)
+            epoch = state.epoch
             num_crops = len(batch[INPUT_KEY])
             student_output = outputs["student"]
             teacher_output = outputs["teacher"]
             loss = self.loss(epoch, num_crops, student_output, teacher_output)
         return outputs, loss
 
-    def train_step(
+    def train_step(  # type: ignore
         self,
         batch_idx: int,
         batch: tensor_dict_type,
-        trainer: Any,
+        state: TrainerState,
+        optimizers: Dict[str, Optimizer],
+        use_amp: bool,
+        grad_scaler: GradScaler,
+        trainer: ITrainer,
         forward_kwargs: Dict[str, Any],
-        loss_kwargs: Dict[str, Any],
     ) -> StepOutputs:
-        state = trainer.state
         if self.lr_schedule is None:
             self.lr_schedule = cosine_scheduler(
                 self.lr * (len(batch[INPUT_KEY][0]) * get_world_size()) / 256.0,  # type: ignore
@@ -419,19 +429,19 @@ class DINO(ModelWithCustomSteps):
                 state.num_step_per_epoch,
             )
         # manual scheduling
-        optimizer = trainer.optimizers["all"]
+        optimizer = optimizers["all"]
         for i, param_group in enumerate(optimizer.param_groups):
             param_group["lr"] = self.lr_schedule[state.step]
             if i == 0:
                 param_group["weight_decay"] = self.wd_schedule[state.step]
         # forward pass
-        rs, loss = self._get_loss(batch_idx, batch, trainer, forward_kwargs)
+        rs, loss = self._get_loss(batch_idx, batch, state, use_amp, forward_kwargs)
         # backward pass
         optimizer.zero_grad()
-        trainer.grad_scaler.scale(loss).backward()
+        grad_scaler.scale(loss).backward()
         # clip norm
         if trainer.clip_norm > 0.0:
-            trainer.grad_scaler.unscale_(optimizer)
+            grad_scaler.unscale_(optimizer)
             nn.utils.clip_grad_norm_(
                 self.student_for_training.parameters(),
                 max_norm=trainer.clip_norm,
@@ -442,8 +452,8 @@ class DINO(ModelWithCustomSteps):
                 if "last_layer" in n:
                     p.grad = None
         # update parameters
-        trainer.grad_scaler.step(optimizer)
-        trainer.grad_scaler.update()
+        grad_scaler.step(optimizer)
+        grad_scaler.update()
         # update momentum teacher
         if self.momentum_schedule is None:
             self.momentum_schedule = cosine_scheduler(
@@ -464,31 +474,22 @@ class DINO(ModelWithCustomSteps):
 
     def evaluate_step(  # type: ignore
         self,
-        loader: CVLoader,
-        portion: float,
-        trainer: Any,
+        batch_idx: int,
+        batch: tensor_dict_type,
+        state: TrainerState,
     ) -> MetricsOutputs:
-        losses = []
-        for i, batch in enumerate(loader):
-            if i / len(loader) >= portion:
-                break
-            batch = to_device(batch, self.device)
-            outputs = self._get_outputs(i, batch, trainer, {})
-            losses.append(
-                self.evaluate_loss(
-                    trainer.state.epoch,
-                    outputs["student"],
-                    outputs["teacher"],
-                )
-            )
-        # gather
-        mean_loss = sum(losses) / len(losses)
+        outputs = self._get_outputs(batch_idx, batch, state, {})
+        loss = self.evaluate_loss(
+            state.epoch,
+            outputs["student"],
+            outputs["teacher"],
+        )
         return MetricsOutputs(
-            -mean_loss,
+            -loss,
             {
-                "loss": mean_loss,
-                "lr": self.lr_schedule[trainer.state.step],  # type: ignore
-                "wd": self.wd_schedule[trainer.state.step],  # type: ignore
+                "loss": loss,
+                "lr": self.lr_schedule[state.step],  # type: ignore
+                "wd": self.wd_schedule[state.step],  # type: ignore
             },
         )
 
@@ -504,9 +505,6 @@ class DINO(ModelWithCustomSteps):
                 regularized.append(param)
         return [{"params": regularized}, {"params": bias_and_norm, "weight_decay": 0.0}]
 
-    def _init_with_trainer(self, trainer: Any) -> None:
-        self.teacher_for_training.requires_grad_(False)
-
     def init_ddp(self) -> None:
         if has_batch_norms(self.student):
             self.student = nn.SyncBatchNorm.convert_sync_batchnorm(self.student)
@@ -515,6 +513,9 @@ class DINO(ModelWithCustomSteps):
         self.ddp_teacher = DDP(self.teacher)
         self.ddp_teacher.requires_grad_(False)  # type: ignore
 
+    def init_with_trainer(self, trainer: ITrainer) -> None:
+        self.teacher_for_training.requires_grad_(False)
+
     def permute_trainer_config(self, trainer_config: Dict[str, Any]) -> None:
         # TODO : make `permute_trainer_config` more general
         if trainer_config["clip_norm"] == 0.0:
