@@ -6,6 +6,7 @@ import numpy as np
 import torch.nn as nn
 
 from abc import abstractmethod
+from abc import ABC
 from abc import ABCMeta
 from copy import deepcopy
 from tqdm import tqdm
@@ -13,28 +14,38 @@ from typing import Any
 from typing import Dict
 from typing import List
 from typing import Type
+from typing import Tuple
 from typing import Union
 from typing import TypeVar
 from typing import Callable
+from typing import Iterator
 from typing import Optional
 from typing import NamedTuple
+from torch.optim import Optimizer
+from cftool.misc import filter_kw
+from cftool.misc import print_info
+from cftool.misc import print_warning
 from cftool.misc import lock_manager
+from cftool.misc import check_requires
 from cftool.misc import shallow_copy_dict
 from cftool.misc import context_error_handler
+from cftool.misc import get_num_positional_args
 from cftool.misc import WithRegister
+from cftool.array import to_numpy
+from cftool.array import to_device
+from cftool.types import np_dict_type
+from cftool.types import tensor_dict_type
+from torch.optim.lr_scheduler import _LRScheduler
 
 from .types import losses_type
-from .types import np_dict_type
-from .types import tensor_dict_type
+from .types import configs_type
+from .types import sample_weights_type
 from .constants import LOSS_KEY
 from .constants import INPUT_KEY
 from .constants import LABEL_KEY
-from .constants import INFO_PREFIX
-from .constants import WARNING_PREFIX
 from .constants import PREDICTIONS_KEY
 from .constants import BATCH_INDICES_KEY
-from .misc.toolkit import to_numpy
-from .misc.toolkit import to_device
+from .constants import ORIGINAL_LABEL_KEY
 from .misc.toolkit import eval_context
 from .misc.toolkit import get_world_size
 from .misc.toolkit import fix_denormal_states
@@ -46,23 +57,35 @@ except:
     onnx = None
 try:
     from onnxsim import simplify as onnx_simplify
-    from onnxsim.onnx_simplifier import get_input_names
+
+    def get_inputs(model: onnx.ModelProto) -> List[onnx.ValueInfoProto]:
+        initializer_names = [x.name for x in model.graph.initializer]
+        return [inp for inp in model.graph.input if inp.name not in initializer_names]
+
+    def get_input_names(model: onnx.ModelProto) -> List[str]:
+        input_names = [inp.name for inp in get_inputs(model)]
+        return input_names
+
 except:
-    onnx_simplify = get_input_names = None
+    onnx_simplify = get_input_names = None  # type: ignore
 
 
-dataset_dict: Dict[str, Type["DatasetProtocol"]] = {}
-loader_dict: Dict[str, Type["DataLoaderProtocol"]] = {}
-model_dict: Dict[str, Type["ModelProtocol"]] = {}
+dataset_dict: Dict[str, Type["IDataset"]] = {}
+loader_dict: Dict[str, Type["IDataLoader"]] = {}
+model_dict: Dict[str, Type["IDLModel"]] = {}
 monitor_dict: Dict[str, Type["TrainerMonitor"]] = {}
-loss_dict: Dict[str, Type["LossProtocol"]] = {}
-metric_dict: Dict[str, Type["MetricProtocol"]] = {}
+loss_dict: Dict[str, Type["ILoss"]] = {}
+multi_prefix_mapping: Dict[str, Type["MultiLoss"]] = {}
+metric_dict: Dict[str, Type["_IMetric"]] = {}
+callback_dict: Dict[str, Type["TrainerCallback"]] = {}
+
+LossType = TypeVar("LossType", bound="ILoss", covariant=True)
 
 
 # data
 
 
-class DatasetProtocol(WithRegister["DatasetProtocol"], metaclass=ABCMeta):
+class IDataset(WithRegister["IDataset"], metaclass=ABCMeta):
     d = dataset_dict
 
     def __init__(self, *args: Any, **kwargs: Any):
@@ -73,16 +96,16 @@ class DatasetProtocol(WithRegister["DatasetProtocol"], metaclass=ABCMeta):
         pass
 
 
-class DataLoaderProtocol(WithRegister["DataLoaderProtocol"], metaclass=ABCMeta):
+class IDataLoader(WithRegister["IDataLoader"], metaclass=ABCMeta):
     d = loader_dict
-    data: DatasetProtocol
+    data: IDataset
     batch_size: int
 
     def __init__(self, *, sample_weights: Optional[np.ndarray] = None):
         self.sample_weights = sample_weights
 
     @abstractmethod
-    def __iter__(self) -> "DataLoaderProtocol":
+    def __iter__(self) -> "IDataLoader":
         pass
 
     @abstractmethod
@@ -100,12 +123,12 @@ class DataLoaderProtocol(WithRegister["DataLoaderProtocol"], metaclass=ABCMeta):
     def __len__(self) -> int:
         return math.ceil(len(self.data) / self.batch_size)
 
-    def copy(self) -> "DataLoaderProtocol":
+    def copy(self) -> "IDataLoader":
         return deepcopy(self)
 
     def temporarily_disable_shuffle(self) -> context_error_handler:
         class _(context_error_handler):
-            def __init__(self, loader: DataLoaderProtocol):
+            def __init__(self, loader: IDataLoader):
                 self.loader = loader
 
             def __enter__(self) -> None:
@@ -117,20 +140,85 @@ class DataLoaderProtocol(WithRegister["DataLoaderProtocol"], metaclass=ABCMeta):
         return _(self)
 
 
+class IDataModule(ABC):
+    id_file: str
+    info_name: str
+    data_folder: str
+    package_folder: str
+
+    @property
+    @abstractmethod
+    def info(self) -> Dict[str, Any]:
+        pass
+
+    @abstractmethod
+    def prepare(self, sample_weights: sample_weights_type) -> None:
+        pass
+
+    @abstractmethod
+    def initialize(self) -> Any:
+        pass
+
+    @abstractmethod
+    def save_info(self, folder: str) -> None:
+        pass
+
+    @abstractmethod
+    def save(self, folder: str) -> None:
+        pass
+
+
 # model
 
 
-class ModelProtocol(nn.Module, WithRegister["ModelProtocol"], metaclass=ABCMeta):
+def _forward(
+    m: nn.Module,
+    batch_idx: int,
+    batch: tensor_dict_type,
+    general_input_key: str,
+    state: Optional["TrainerState"] = None,
+    *,
+    general_output_key: str = PREDICTIONS_KEY,
+    **kwargs: Any,
+) -> tensor_dict_type:
+    fn = m.forward
+    if check_requires(fn, "general_output_key"):
+        kwargs["general_output_key"] = general_output_key
+    kw = filter_kw(fn, kwargs)
+    args: List[Any] = []
+    if check_requires(fn, "batch_idx"):
+        args.append(batch_idx)
+    if get_num_positional_args(fn) > 0:
+        args.append(batch if check_requires(fn, "batch") else batch[general_input_key])
+    if check_requires(fn, "state"):
+        args.append(state)
+    rs = m(*args, **kw)
+    if not isinstance(rs, dict):
+        rs = {general_output_key: rs}
+    return rs
+
+
+class WithDeviceMixin:
+    parameters: Callable[["WithDeviceMixin"], Iterator[nn.Parameter]]
+
+    @property
+    def device(self) -> torch.device:
+        params = list(self.parameters())
+        return torch.device("cpu") if not params else params[0].device
+
+
+class IDLModel(
+    WithDeviceMixin,
+    nn.Module,
+    WithRegister["IDLModel"],
+    metaclass=ABCMeta,
+):
     d = model_dict
 
     def __init__(self, *args: Any, **kwargs: Any):
         super().__init__()
 
-    @property
-    def device(self) -> torch.device:
-        return list(self.parameters())[0].device
-
-    def _init_with_trainer(self, trainer: Any) -> None:
+    def _init_with_trainer(self, trainer: "ITrainer") -> None:
         pass
 
     @abstractmethod
@@ -163,7 +251,7 @@ class ModelProtocol(nn.Module, WithRegister["ModelProtocol"], metaclass=ABCMeta)
         num_samples: Optional[int] = None,
         verbose: bool = True,
         **kwargs: Any,
-    ) -> "ModelProtocol":
+    ) -> "IDLModel":
         # prepare
         device = self.device
         model = self.cpu()
@@ -232,38 +320,36 @@ class ModelProtocol(nn.Module, WithRegister["ModelProtocol"], metaclass=ABCMeta)
                 if not simplify:
                     return self.to(device)
                 if onnx is None:
-                    print(
-                        f"{WARNING_PREFIX}`onnx` is not installed, "
-                        f"so the exported onnx model will not be simplified"
+                    print_warning(
+                        "`onnx` is not installed, "
+                        "so the exported onnx model will not be simplified"
                     )
                     return self.to(device)
                 if onnx_simplify is None or get_input_names is None:
-                    print(
-                        f"{WARNING_PREFIX}`onnx-simplifier` is not installed, "
-                        f"so the exported onnx model will not be simplified"
+                    print_warning(
+                        "`onnx-simplifier` is not installed, "
+                        "so the exported onnx model will not be simplified"
                     )
                     return self.to(device)
                 try:
                     onnx_model = onnx.load(onnx_path)
                     final_input_names = get_input_names(onnx_model)
-                    np_sample = {
-                        name: to_numpy(tensor)
-                        for name, tensor in input_sample.items()
-                        if name in final_input_names
-                    }
                     model_simplified, check = onnx_simplify(
                         onnx_model,
-                        input_data=np_sample,
-                        dynamic_input_shape=bool(dynamic_axes),
+                        test_input_shapes={
+                            name: tensor.shape
+                            for name, tensor in input_sample.items()
+                            if name in final_input_names
+                        },
                     )
                 except Exception as err:
                     if verbose:
-                        print(f"{WARNING_PREFIX}Failed to simplify ONNX model ({err})")
+                        print_warning(f"Failed to simplify ONNX model ({err})")
                     model_simplified = None
                     check = False
                 if verbose:
                     tag = " " if check else " not "
-                    print(f"{INFO_PREFIX}Simplified ONNX model is{tag}validated!")
+                    print_info(f"Simplified ONNX model is{tag}validated!")
                 if check and model_simplified is not None:
                     onnx.save(model_simplified, onnx_path)
         return self.to(device)
@@ -286,7 +372,7 @@ class InferenceOutputs(NamedTuple):
     loss_items: Optional[Dict[str, float]]
 
 
-class ModelWithCustomSteps(ModelProtocol, metaclass=ABCMeta):
+class ModelWithCustomSteps(IDLModel, metaclass=ABCMeta):
     custom_train_step: bool = True
     custom_evaluate_step: bool = True
     custom_params_groups: bool = False
@@ -297,18 +383,19 @@ class ModelWithCustomSteps(ModelProtocol, metaclass=ABCMeta):
         self,
         batch_idx: int,
         batch: tensor_dict_type,
-        trainer: Any,
+        trainer: "ITrainer",
         forward_kwargs: Dict[str, Any],
         loss_kwargs: Dict[str, Any],
     ) -> StepOutputs:
         pass
 
+    # TODO : Add `forward_kwargs`
     @abstractmethod
     def evaluate_step(
         self,
-        loader: DataLoaderProtocol,
+        loader: IDataLoader,
         portion: float,
-        trainer: Any,
+        trainer: "ITrainer",
     ) -> MetricsOutputs:
         pass
 
@@ -322,13 +409,13 @@ class ModelWithCustomSteps(ModelProtocol, metaclass=ABCMeta):
         pass
 
 
-# trainer
+# trainer types
 
 
 class TrainerState:
     def __init__(
         self,
-        loader: DataLoaderProtocol,
+        loader: IDataLoader,
         *,
         num_epoch: int,
         max_epoch: int,
@@ -342,6 +429,7 @@ class TrainerState:
         num_step_per_log: int = 350,
         num_step_per_snapshot: Optional[int] = None,
         max_step_per_snapshot: int = 1000,
+        min_snapshot_epoch_gap: int = 0,
     ):
         self.step = self.epoch = 0
         self.batch_size = loader.batch_size * get_world_size()
@@ -363,10 +451,15 @@ class TrainerState:
             num_step_per_snapshot = min(max_step_per_snapshot, num_step_per_snapshot)
         self.num_step_per_snapshot = num_step_per_snapshot
         self.max_step_per_snapshot = max_step_per_snapshot
+        self.min_snapshot_epoch_gap = min_snapshot_epoch_gap
+        self._previous_snapshot_epoch = 0
 
     def set_terminate(self) -> None:
         self.step = self.epoch = -1
 
+    def update_snapshot_epoch(self) -> None:
+        self._previous_snapshot_epoch = self.epoch
+
     @property
     def config(self) -> Dict[str, Any]:
         return {
@@ -433,6 +526,12 @@ class TrainerState:
         period = max(1, int(min_period)) * self.num_step_per_snapshot
         return self.step % period == 0
 
+    @property
+    def can_snapshot(self) -> bool:
+        if self.is_terminate:
+            return True
+        return self.epoch - self._previous_snapshot_epoch >= self.min_snapshot_epoch_gap
+
     @property
     def should_start_snapshot(self) -> bool:
         return self.step >= self.snapshot_start_step
@@ -495,10 +594,7 @@ class MonitorResults(NamedTuple):
 # loss
 
 
-LossType = TypeVar("LossType", bound="LossProtocol", covariant=True)
-
-
-class LossProtocol(nn.Module, WithRegister[LossType], metaclass=ABCMeta):
+class ILoss(nn.Module, WithRegister[LossType], metaclass=ABCMeta):
     d = loss_dict
     placeholder_key = "[PLACEHOLDER]"
 
@@ -559,23 +655,19 @@ class LossProtocol(nn.Module, WithRegister[LossType], metaclass=ABCMeta):
         return AuxLoss.register_(loss_name, aux_names)
 
 
-multi_prefix_mapping: Dict[str, Type["MultiLoss"]] = {}
-
-
-class MultiLoss(LossProtocol, metaclass=ABCMeta):
+class MultiLoss(ILoss, metaclass=ABCMeta):
     prefix: str
 
     names: Union[str, List[str]]
     base_losses: nn.ModuleList
 
     def _init_config(self) -> None:
-        base_losses: List[LossProtocol]
+        base_losses: List[ILoss]
         if isinstance(self.names, str):
-            base_losses = [LossProtocol.make(self.names, self.config)]
+            base_losses = [ILoss.make(self.names, self.config)]
         else:
             base_losses = [
-                LossProtocol.make(name, self.config.get(name, {}))
-                for name in self.names
+                ILoss.make(name, self.config.get(name, {})) for name in self.names
             ]
         self.base_losses = nn.ModuleList(base_losses)
 
@@ -616,16 +708,16 @@ class MultiLoss(LossProtocol, metaclass=ABCMeta):
         return _
 
 
-class AuxLoss(LossProtocol):
+class AuxLoss(ILoss):
     identifier: str = "aux"
     main_loss_key: str = "main"
 
-    loss: LossProtocol
+    loss: ILoss
     loss_name: str
     aux_names: List[str]
 
     def _init_config(self) -> None:
-        self.loss = LossProtocol.make(self.loss_name, self.config)
+        self.loss = ILoss.make(self.loss_name, self.config)
 
     @staticmethod
     def _convert(losses: losses_type) -> torch.Tensor:
@@ -678,8 +770,8 @@ class AuxLoss(LossProtocol):
         return tag
 
 
-@LossProtocol.register(LossProtocol.placeholder_key)
-class PlaceholderLoss(LossProtocol):
+@ILoss.register(ILoss.placeholder_key)
+class PlaceholderLoss(ILoss):
     def _core(
         self,
         forward_results: tensor_dict_type,
@@ -693,12 +785,12 @@ class PlaceholderLoss(LossProtocol):
 # inference
 
 
-class InferenceProtocol:
+class IInference:
     def __init__(
         self,
         *,
         onnx: Optional[ONNX] = None,
-        model: Optional[ModelProtocol] = None,
+        model: Optional[IDLModel] = None,
         use_grad_in_predict: bool = False,
     ):
         self.onnx = onnx
@@ -711,12 +803,12 @@ class InferenceProtocol:
 
     def get_outputs(
         self,
-        loader: DataLoaderProtocol,
+        loader: IDataLoader,
         *,
         portion: float = 1.0,
         state: Optional[TrainerState] = None,
-        metrics: Optional["MetricProtocol"] = None,
-        loss: Optional[LossProtocol] = None,
+        metrics: Optional["_IMetric"] = None,
+        loss: Optional[ILoss] = None,
         return_outputs: bool = True,
         stack_outputs: bool = True,
         use_tqdm: bool = False,
@@ -782,7 +874,11 @@ class InferenceProtocol:
                         results.setdefault(k, []).append(v_np)  # type: ignore
                 if requires_np:
                     for k, v in batch.items():
-                        if k == INPUT_KEY or k.endswith(BATCH_INDICES_KEY):
+                        if (
+                            k == INPUT_KEY
+                            or k == ORIGINAL_LABEL_KEY
+                            or k.endswith(BATCH_INDICES_KEY)
+                        ):
                             continue
                         if v is None:
                             continue
@@ -793,7 +889,7 @@ class InferenceProtocol:
                         labels.setdefault(k, []).append(v)
                 # metrics
                 if requires_metrics:
-                    sub_outputs = metrics.evaluate(np_batch, np_outputs)  # type: ignore
+                    sub_outputs = metrics.evaluate(np_batch, np_outputs, loader)  # type: ignore
                     metric_outputs_list.append(sub_outputs)
                 # loss
                 if loss is not None:
@@ -825,6 +921,7 @@ class InferenceProtocol:
                 metric_outputs = metrics.evaluate(
                     {k: np.vstack(v) for k, v in labels.items()},
                     final_results,
+                    loader,
                 )
             else:
                 scores = []
@@ -860,10 +957,10 @@ class InferenceProtocol:
 # metrics
 
 
-class MetricProtocol(WithRegister["MetricProtocol"], metaclass=ABCMeta):
+class _IMetric(WithRegister["_IMetric"], metaclass=ABCMeta):
     d = metric_dict
 
-    trainer: Any
+    trainer: "ITrainer"
 
     def __init__(self, *args: Any, **kwargs: Any):
         pass
@@ -878,7 +975,7 @@ class MetricProtocol(WithRegister["MetricProtocol"], metaclass=ABCMeta):
         self,
         np_batch: np_dict_type,
         np_outputs: np_dict_type,
-        loader: Optional[DataLoaderProtocol],
+        loader: Optional[IDataLoader],
     ) -> float:
         pass
 
@@ -886,35 +983,378 @@ class MetricProtocol(WithRegister["MetricProtocol"], metaclass=ABCMeta):
     def requires_all(self) -> bool:
         return False
 
+    @classmethod
+    def fuse(
+        cls,
+        names: Union[str, List[str]],
+        configs: configs_type = None,
+        *,
+        metric_weights: Optional[Dict[str, float]] = None,
+    ) -> "_IMetric":
+        metrics = _IMetric.make_multiple(names, configs)
+        if isinstance(metrics, _IMetric):
+            return metrics
+        return _MultipleMetrics(metrics, weights=metric_weights)
+
     def evaluate(
         self,
         np_batch: np_dict_type,
         np_outputs: np_dict_type,
-        loader: Optional[DataLoaderProtocol] = None,
+        loader: Optional[IDataLoader] = None,
     ) -> MetricsOutputs:
         metric = self._core(np_batch, np_outputs, loader)
         score = metric * (1.0 if self.is_positive else -1.0)
         return MetricsOutputs(score, {self.__identifier__: metric})
 
 
+class _MultipleMetrics(_IMetric):
+    @property
+    def is_positive(self) -> bool:
+        raise NotImplementedError
+
+    @property
+    def requires_all(self) -> bool:
+        return any(metric.requires_all for metric in self.metrics)
+
+    def _core(
+        self,
+        np_batch: np_dict_type,
+        np_outputs: np_dict_type,
+        loader: Optional[IDataLoader],
+    ) -> float:
+        raise NotImplementedError
+
+    def __init__(
+        self,
+        metric_list: List[_IMetric],
+        *,
+        weights: Optional[Dict[str, float]] = None,
+    ):
+        super().__init__()
+        self.metrics = metric_list
+        self.weights = weights or {}
+        self.__identifier__ = " | ".join(m.__identifier__ for m in metric_list)
+
+    def evaluate(
+        self,
+        np_batch: np_dict_type,
+        np_outputs: np_dict_type,
+        loader: Optional[IDataLoader] = None,
+    ) -> MetricsOutputs:
+        scores: List[float] = []
+        weights: List[float] = []
+        metrics_values: Dict[str, float] = {}
+        for metric in self.metrics:
+            metric_outputs = metric.evaluate(np_batch, np_outputs, loader)
+            w = self.weights.get(metric.__identifier__, 1.0)
+            weights.append(w)
+            scores.append(metric_outputs.final_score * w)
+            metrics_values.update(metric_outputs.metric_values)
+        return MetricsOutputs(sum(scores) / sum(weights), metrics_values)
+
+
+# trainer interface
+
+
+class DeviceInfo(NamedTuple):
+    cuda: Optional[str]
+    rank: Optional[int]
+
+    @property
+    def device(self) -> torch.device:
+        if self.rank is not None:
+            return torch.device(f"cuda:{self.rank}")
+        return torch.device("cpu" if self.cuda is None else f"cuda:{self.cuda}")
+
+
+class OptimizerPack(NamedTuple):
+    scope: str
+    optimizer_name: str
+    scheduler_name: Optional[str] = None
+    optimizer_config: Optional[Dict[str, Any]] = None
+    scheduler_config: Optional[Dict[str, Any]] = None
+
+
+class TqdmSettings(NamedTuple):
+    use_tqdm: bool = False
+    use_step_tqdm: bool = False
+    use_tqdm_in_validation: bool = False
+    in_distributed: bool = False
+    position: int = 0
+    desc: str = "epoch"
+
+
+class TrainerCallback(WithRegister["TrainerCallback"]):
+    d = callback_dict
+    is_rank_0: bool = True
+
+    def __init__(self, *args: Any, **kwargs: Any):
+        pass
+
+    def initialize(self) -> None:
+        pass
+
+    def mutate_train_forward_kwargs(
+        self,
+        kwargs: Dict[str, Any],
+        trainer: "ITrainer",
+    ) -> None:
+        pass
+
+    def mutate_train_loss_kwargs(
+        self,
+        kwargs: Dict[str, Any],
+        trainer: "ITrainer",
+    ) -> None:
+        pass
+
+    def before_loop(self, trainer: "ITrainer") -> None:
+        pass
+
+    def log_lr(self, key: str, lr: float, state: TrainerState) -> None:
+        pass
+
+    def log_metrics(self, metric_outputs: MetricsOutputs, state: TrainerState) -> None:
+        pass
+
+    def log_metrics_msg(
+        self,
+        metric_outputs: MetricsOutputs,
+        metrics_log_path: str,
+        state: TrainerState,
+    ) -> None:
+        pass
+
+    def log_artifacts(self, trainer: "ITrainer") -> None:
+        pass
+
+    def after_step(self, step_outputs: StepOutputs, state: TrainerState) -> None:
+        pass
+
+    def after_monitor(
+        self,
+        monitor_results: MonitorResults,
+        state: TrainerState,
+    ) -> None:
+        pass
+
+    def finalize(self, trainer: "ITrainer") -> None:
+        pass
+
+
+class ITrainer(ABC):
+    loss: ILoss
+    model: IDLModel
+    metrics: Optional[_IMetric]
+    monitors: List[TrainerMonitor]
+    callbacks: List[TrainerCallback]
+    optimizers: Dict[str, Optimizer]
+    schedulers: Dict[str, Optional[_LRScheduler]]
+
+    state: TrainerState
+    device_info: DeviceInfo
+    train_loader: IDataLoader
+    train_loader_copy: IDataLoader
+    valid_loader: Optional[IDataLoader]
+    inference: IInference
+
+    workplace: str
+    checkpoint_folder: Optional[str]
+
+    tqdm_settings: TqdmSettings
+    state_config: Dict[str, Any]
+    num_epoch: int
+    max_epoch: int
+    fixed_steps: Optional[int]
+    valid_portion: float
+    use_amp: bool
+    use_zero: bool
+    clip_norm: float
+    grad_scaler: torch.cuda.amp.GradScaler
+
+    @property
+    @abstractmethod
+    def config(self) -> Dict[str, Any]:
+        pass
+
+    @property
+    @abstractmethod
+    def device(self) -> torch.device:
+        pass
+
+    @property
+    @abstractmethod
+    def use_tqdm_in_validation(self) -> bool:
+        pass
+
+    @property
+    @abstractmethod
+    def validation_loader(self) -> IDataLoader:
+        pass
+
+    @property
+    @abstractmethod
+    def input_sample(self) -> tensor_dict_type:
+        pass
+
+    @property
+    @abstractmethod
+    def has_checkpoint_folder(self) -> bool:
+        pass
+
+    @property
+    @abstractmethod
+    def model_has_custom_steps(self) -> bool:
+        pass
+
+    @property
+    @abstractmethod
+    def model_for_training(self) -> nn.Module:
+        pass
+
+    # init
+
+    @abstractmethod
+    def _init_ddp(self) -> None:
+        pass
+
+    @abstractmethod
+    def _init_finetune(self) -> None:
+        pass
+
+    @abstractmethod
+    def default_lr_configs(
+        self,
+        optimizer: Optimizer,
+        optimizer_config: Dict[str, Any],
+    ) -> Dict[str, Dict[str, Any]]:
+        pass
+
+    @abstractmethod
+    def _define_optimizer(self, pack: OptimizerPack) -> Optimizer:
+        pass
+
+    @abstractmethod
+    def _define_scheduler(self, optimizer: Optimizer, pack: OptimizerPack) -> None:
+        pass
+
+    @abstractmethod
+    def _init_optimizers(self) -> None:
+        pass
+
+    # core
+
+    @abstractmethod
+    def post_loss_step(self, loss_dict: tensor_dict_type) -> None:
+        pass
+
+    @abstractmethod
+    def weighted_loss_score(self, loss_items: Dict[str, float]) -> float:
+        pass
+
+    @abstractmethod
+    def clip_norm_step(self) -> None:
+        pass
+
+    @abstractmethod
+    def optimizer_step(self) -> None:
+        pass
+
+    @abstractmethod
+    def scheduler_step(self) -> None:
+        pass
+
+    @abstractmethod
+    def _get_scheduler_settings(
+        self,
+        key: str,
+        scheduler: Any,
+    ) -> Tuple[bool, Dict[str, Any]]:
+        pass
+
+    @abstractmethod
+    def _logging_step(self, metrics_outputs: MetricsOutputs) -> None:
+        pass
+
+    @abstractmethod
+    def _monitor_step(self, step_outputs: StepOutputs) -> MonitorResults:
+        pass
+
+    @abstractmethod
+    def _step(self, batch_idx: int, batch: tensor_dict_type) -> StepOutputs:
+        pass
+
+    @abstractmethod
+    def fit(
+        self,
+        data: IDataModule,
+        loss: ILoss,
+        model: IDLModel,
+        inference: IInference,
+        *,
+        config_export_file: Optional[str] = None,
+        show_summary: Optional[bool] = None,
+        cuda: Optional[str] = None,
+    ) -> "ITrainer":
+        pass
+
+    @abstractmethod
+    def get_metrics(
+        self,
+        *,
+        portion: float = 1.0,
+        loader: Optional[IDataLoader] = None,
+    ) -> MetricsOutputs:
+        pass
+
+    @abstractmethod
+    def save_checkpoint(
+        self,
+        score: float,
+        folder: Optional[str] = None,
+        *,
+        no_history: bool = False,
+    ) -> None:
+        pass
+
+    @abstractmethod
+    def restore_checkpoint(
+        self,
+        folder: str = None,
+        strict: bool = True,
+        state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
+    ) -> bool:
+        pass
+
+
 __all__ = [
+    "_forward",
     "dataset_dict",
     "loader_dict",
     "loss_dict",
-    "DatasetProtocol",
-    "DataLoaderProtocol",
-    "ModelProtocol",
+    "multi_prefix_mapping",
+    "metric_dict",
+    "callback_dict",
+    "IDataset",
+    "IDataLoader",
+    "IDLModel",
     "ModelWithCustomSteps",
     "StepOutputs",
     "TrainerState",
     "TrainerMonitor",
     "MonitorResults",
-    "LossProtocol",
+    "ILoss",
     "MultiLoss",
     "AuxLoss",
     "ONNX",
     "InferenceOutputs",
-    "InferenceProtocol",
+    "IInference",
     "MetricsOutputs",
-    "MetricProtocol",
+    "_IMetric",
+    "_MultipleMetrics",
+    "DeviceInfo",
+    "OptimizerPack",
+    "TqdmSettings",
+    "TrainerCallback",
+    "ITrainer",
 ]
